import asyncio
import itertools
import logging
import queue
import threading
import time
import uuid
from contextlib import suppress

import aiohttp
import pygame
import uvloop

from pongy import settings
from pongy.models import MoveDirection
from pongy.models import WsCommand
from pongy.models import WsCommandMovePayload
from pongy.models import WsEvent
from pongy.ui.ball import Ball
from pongy.ui.racket import FirstRacket
from pongy.ui.racket import MyRacket
from pongy.ui.racket import SecondRacket
from pongy.ui.racket import ThirdRacket

logger = logging.getLogger(__name__)


class Exit:
    pass


def run_app(host: str, port: int):
    pygame.init()
    quit = False
    input_queue: queue.Queue = queue.Queue()
    output_queue: queue.Queue = queue.Queue()
    player_id = str(uuid.uuid4())

    def ws_connection():
        async def _process_output_queue(ws):
            while True:
                await asyncio.sleep(0.01)
                with suppress(queue.Empty):
                    await ws.send_json(output_queue.get_nowait().dict())

        async def _handle():
            url = f"ws://{host}:{port}/ws"
            try:
                async with aiohttp.ClientSession() as session:
                    async with session.ws_connect(
                        url,
                        heartbeat=settings.WS_HEARTBEAT_TIMEOUT,
                        headers={"Cookie": f"player_id={player_id}"},
                    ) as ws:
                        task = asyncio.ensure_future(_process_output_queue(ws))
                        async for msg in ws:
                            if quit or msg.type == aiohttp.WSMsgType.ERROR:
                                task.cancel()
                                break
                            if msg.type == aiohttp.WSMsgType.TEXT:
                                ws_event = WsEvent.parse_raw(msg.data)
                                input_queue.put(ws_event)
            except aiohttp.ClientConnectionError:
                logger.error("Connection error")
                input_queue.put(Exit())

        uvloop.install()
        loop = asyncio.new_event_loop()
        loop.run_until_complete(_handle())

    keep_connection = threading.Thread(target=ws_connection)
    keep_connection.start()

    surface = pygame.display.set_mode(settings.BOARD_SIZE)
    surface.fill(settings.BOARD_COLOR)

    while not quit:
        time.sleep(0.01)
        while not input_queue.empty():
            ws_event = input_queue.get_nowait()
            if isinstance(ws_event, Exit):
                quit = True
                break
            surface.fill(settings.BOARD_COLOR)
            for player, racket in zip(
                itertools.chain(
                    (p for p in ws_event.data.payload.players if p.uuid == player_id),
                    (p for p in ws_event.data.payload.players if p.uuid != player_id),
                ),
                (MyRacket, FirstRacket, SecondRacket, ThirdRacket),
            ):
                racket(player.racket_position).draw(surface)

            ball_position = ws_event.data.payload.ball_position
            Ball(ball_position).draw(surface)
            pygame.display.flip()
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                quit = True
            elif event.type == pygame.KEYDOWN:
                if event.key == pygame.K_q:
                    quit = True
        keys = pygame.key.get_pressed()
        if keys[pygame.K_LEFT]:
            ws_event = WsCommand(
                payload=WsCommandMovePayload(direction=MoveDirection.LEFT)
            )
            output_queue.put(ws_event)
        if keys[pygame.K_RIGHT]:
            ws_event = WsCommand(
                payload=WsCommandMovePayload(direction=MoveDirection.RIGHT)
            )
            output_queue.put(ws_event)
    pygame.quit()



import pygame

from pongy import settings


class BaseRacket:
    def __init__(self, position):
        self._position = position

    def draw(self, surface):
        pass


class HorizontalRacket(BaseRacket):
    def __init__(self, position):
        super().__init__(position)
        self._width = settings.RACKET_LENGTH
        self._height = settings.RACKET_HEIGHT


class VerticalRacket(BaseRacket):
    def __init__(self, position):
        super().__init__(position)
        self._width = settings.RACKET_HEIGHT
        self._height = settings.RACKET_LENGTH


class FirstRacket(HorizontalRacket):
    def draw(self, surface):
        pygame.draw.rect(
            surface,
            settings.RACKET_COLOR,
            pygame.Rect(
                self._position,
                0,
                self._width,
                self._height,
            ),
        )


class SecondRacket(VerticalRacket):
    def draw(self, surface):
        pygame.draw.rect(
            surface,
            settings.RACKET_COLOR,
            pygame.Rect(
                0,
                self._position,
                self._width,
                self._height,
            ),
        )


class ThirdRacket(VerticalRacket):
    def draw(self, surface):
        pygame.draw.rect(
            surface,
            settings.RACKET_COLOR,
            pygame.Rect(
                settings.BOARD_SIZE[0] - settings.RACKET_HEIGHT,
                self._position,
                self._width,
                self._height,
            ),
        )


class MyRacket(HorizontalRacket):
    def draw(self, surface):
        pygame.draw.rect(
            surface,
            settings.RACKET_COLOR,
            pygame.Rect(
                self._position,
                settings.BOARD_SIZE[0] - settings.RACKET_HEIGHT,
                self._width,
                self._height,
            ),
        )



import asyncio
import logging
import math
from random import randint
from types import TracebackType

from aiohttp import web

from pongy import settings
from pongy.models import MoveDirection
from pongy.models import WsEvent
from pongy.models import WsGameStateEvent
from pongy.models import WsGameStatePayload
from pongy.models import WsPlayer


logger = logging.getLogger(__name__)


class Player:
    def __init__(self, uuid: str, ws: web.WebSocketResponse):
        self.uuid: str = uuid
        self.ws: web.WebSocketResponse = ws
        self.score: int = 0
        self.racket_position: int = 325

    def move_racket(self, direction: MoveDirection) -> None:
        if (
            direction == MoveDirection.LEFT
            and self.racket_position - settings.RACKET_SPEED >= 0
        ):
            self.racket_position -= settings.RACKET_SPEED
        if (
            direction == MoveDirection.RIGHT
            and self.racket_position + settings.RACKET_SPEED
            <= settings.BOARD_SIZE[0] - settings.RACKET_LENGTH
        ):
            self.racket_position += settings.RACKET_SPEED


class Game:
    def __init__(self) -> None:
        self.players: list[Player] = []
        self.ball_position: tuple[int, int] = (345, 345)
        self.ball_angle: int = randint(0, 360)
        self._broadcaster = asyncio.ensure_future(self.broadcast())

    def add_player(self, player: Player) -> None:
        self.players.append(player)

    def remove_player(self, player: Player) -> None:
        self.players[:] = [p for p in self.players if p.uuid != player.uuid]
        if not self.players:
            self._broadcaster.cancel()

    def move_ball(self) -> None:
        new_x = self.ball_position[0] + settings.BALL_SPEED * math.cos(
            math.radians(self.ball_angle)
        )
        new_y = self.ball_position[1] + settings.BALL_SPEED * math.sin(
            math.radians(self.ball_angle)
        )
        if new_x < 0:
            new_x = 0
            self.ball_angle = 180 - self.ball_angle
        elif new_x > settings.BOARD_SIZE[0] - 10:
            new_x = settings.BOARD_SIZE[0] - 10
            self.ball_angle = 180 - self.ball_angle
        if new_y < 0:
            new_y = 0
            self.ball_angle = -self.ball_angle
        elif new_y > settings.BOARD_SIZE[0] - 10:
            new_y = settings.BOARD_SIZE[0] - 10
            self.ball_angle = -self.ball_angle
        self.ball_position = int(new_x), int(new_y)

    def to_payload(self) -> WsGameStatePayload:
        return WsGameStatePayload(
            ball_position=self.ball_position,
            players=[
                WsPlayer(
                    uuid=player.uuid,
                    score=player.score,
                    racket_position=player.racket_position,
                )
                for player in self.players
            ],
        )

    async def broadcast(self):
        while True:
            await self.publish_state()
            await asyncio.sleep(0.05)
            self.move_ball()

    async def publish_state(self) -> None:
        payload = WsEvent(data=WsGameStateEvent(payload=self.to_payload()))
        for subscriber in self.players:
            try:
                await subscriber.ws.send_json(payload.dict())
            except ConnectionResetError as err:
                logger.warning(err)


class GamePool:
    _awaiting: Game | None = None

    def __init__(self, player: Player) -> None:
        self._player: Player = player
        self._game: Game | None = None

    async def __aenter__(self) -> Game:
        if not GamePool._awaiting:
            self._game = GamePool._awaiting = Game()
        else:
            self._game = GamePool._awaiting
        self._game.add_player(self._player)
        if len(self._game.players) == 4:
            GamePool._awaiting = None
        return self._game

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        if self._game:
            self._game.remove_player(self._player)
            if GamePool._awaiting is self._game and not self._game.players:
                GamePool._awaiting = None



import logging


RACKET_COLOR = (200, 0, 0)

BALL_COLOR = (173, 255, 47)

BOARD_COLOR = (233, 150, 122)

BOARD_SIZE = (700, 700)

RACKET_LENGTH = 50

RACKET_HEIGHT = 10

BALL_SPEED = 10

RACKET_SPEED = 5

WS_HEARTBEAT_TIMEOUT = 10

LOGGING_LEVEL = logging.DEBUG

SERVER_HOST = "0.0.0.0"

SERVER_PORT = 8888
